import datetime
from collections import OrderedDict
from collections import defaultdict
from functools import partial

from gisdo import constants as const
from gisdo import xml_helpers
from gisdo.exceptions import NoDataInMo
from gisdo.index import get_enrolled_age_filter
from gisdo.index.direct import ind_19_3
from gisdo.index.direct import ind_29_1
from gisdo.index.enrolled import Index
from gisdo.service import FilialGroupData
from gisdo.service import GroupData
from gisdo.service import ParentGroupData
from gisdo.utils import DateMixin
from kinder.core.children.models import Children
from kinder.core.group.models import Group
from kinder.core.group.models import GroupStatusEnum

from ..collector import DataCollector
from .constants import *
from .helpers import prep_ind_19_3_queries
from .helpers import prep_ind_29_1_queries
from .helpers import prepare_enrolled_index_queries
from .helpers import prep_ind_enrolled_data
from .helpers import ReportQuery


class ReportDataCollector(DataCollector):
    """Сборщик данных отчета "Выгрузка детей по показателю/тегу"."""

    def __init__(self, mo, unit_helper, report_start_date, indexes=None):
        super().__init__(mo, unit_helper, report_start_date)

        self.indexes = indexes or []

    def collect(self, report_form=None):

        # Признак наличия садов по которым можно собрать данные
        dou_list = self._unit_helper.get_mo_units(self._mo.id)
        have_data = len(dou_list)

        if not have_data:
            raise NoDataInMo(
                f'Нет садов в МО {self._mo} пригодных для отправки'
            )

        # Собираем по ДОУ
        for dou in dou_list:
            self._collect_enrolled_data(dou)

        return self._cache[self.ENROLLED]

    @staticmethod
    def _merge_dicts(first_dict, second_dict):
        if not first_dict or not second_dict:
            return

        first_dou_id = list(first_dict.keys())[0]
        second_dou_id = list(second_dict.keys())[0]

        for index, first_index_dict in first_dict[first_dou_id].items():
            second_index_dict = second_dict[second_dou_id][index]
            age_categories = set(first_index_dict.keys()) | set(
                second_index_dict.keys())

            for age_cat in age_categories:
                if age_cat not in second_index_dict:
                    continue

                data = second_index_dict[age_cat]
                if age_cat in first_index_dict:
                    first_index_dict[age_cat].extend(data)
                else:
                    first_index_dict[age_cat] = data.copy()

    def get_enrolled_data(self, collection, age_filter):
        """Фильтрация данных по тегу"""

        data = [data for data in collection if age_filter(*data)]

        return data

    def get_enrolled_index_data(self, _unit, report_start_date):
        """Получение данных для тега enrolled"""

        full_enrolled_data = defaultdict(list)

        filials = xml_helpers.get_dou_filials(
            _unit
        ).select_related("filial")

        parent_group_data = ParentGroupData(
            _unit, report_start_date, filials,
            prep_enrollments_func=prep_ind_enrolled_data
        )

        # головная организация передается только если у нее есть группы
        if xml_helpers.has_fact_groups(_unit):
            for parent_group in Group.objects.filter(
                    status=GroupStatusEnum.FACT, unit=_unit).order_by('id'):
                age_from, age_to = GroupData.get_age_range(parent_group)
                age_cat = f'от {int(age_from)} до {int(age_to)} лет'

                enrolled = parent_group_data.get_enrollments_count(
                    parent_group,
                    prep_enrollments_func=prep_ind_enrolled_data)

                full_enrolled_data[age_cat].extend(enrolled)

        for filial in (fd.filial for fd in filials.iterator()):
            # филиал передается только если у него есть группы
            if xml_helpers.has_fact_groups(filial):

                filial_group_data = FilialGroupData(
                    filial, parent_group_data, report_start_date,
                    prep_enrollments_func=prep_ind_enrolled_data
                )

                for filial_group in Group.objects.filter(
                        status=GroupStatusEnum.FACT, unit=_unit).order_by('id'):
                    age_from, age_to = GroupData.get_age_range(filial_group)
                    age_cat = f'от {int(age_from)} до {int(age_to)} лет'

                    enrolled = filial_group_data.get_enrollments_count(
                        filial_group,
                        prep_enrollments_func=prep_ind_enrolled_data)

                    full_enrolled_data[age_cat].extend(enrolled)

        return full_enrolled_data

    def has_selected_index(self, indexes):
        """Проверяет, был ли выбран хотя бы один из указанных индексов

        :param indexes: Названия индексов для проверки
        :type indexes: Iterable

        :return: Был ли выбран хотя бы один из указанных индексов для отчёта
        :rtype: bool
        """
        return any(index in self.indexes for index in indexes)

    def _collect_enrolled_data(self, dou):

        def _inner(_dou, report_start_date, calc_filial=False):
            enrolled = {
                _dou.id: OrderedDict([
                    (INDEX_ENROLLED, {}),
                    (INDEX_19, {}),
                    (INDEX_19_1, {}),
                    (INDEX_19_2, {}),
                    (INDEX_19_3, {}),
                    (INDEX_29, {}),
                    (INDEX_29_1, {}),
                ])
            }
            dou_enrolled = enrolled[_dou.id]

            if INDEX_ENROLLED in self.indexes:
                ind_enrolled_data = self.get_enrolled_index_data(
                    _dou, report_start_date)
                if ind_enrolled_data:
                    dou_enrolled[INDEX_ENROLLED] = ind_enrolled_data

            enrolled_index = Index(_dou, report_start_date)
            prepared_enrolled_index = partial(
                enrolled_index,
                prep_enrolled_func=prepare_enrolled_index_queries,
            )

            if INDEX_29_1 in self.indexes:
                index_29_1 = ind_29_1(_dou, self._report_start_date,
                                      prep_func=prep_ind_29_1_queries)

            if not calc_filial:
                if self.has_selected_index([INDEX_19, INDEX_19_1, INDEX_19_2]):
                    index_19 = list(prepared_enrolled_index())

            for age_cat in list(const.AGE_CATEGORIES_FULL.values()):
                current_year_enrolled_filter = get_enrolled_age_filter(
                    age_cat, DateMixin.get_current_calendar_year())
                on_date_enrolled_filter = get_enrolled_age_filter(
                    age_cat, datetime.date.today)
                next_calendar_year_enrolled = get_enrolled_age_filter(
                    age_cat, DateMixin.get_next_calendar_year)

                # Общая численность детей зачисленных в ДОО
                if not calc_filial:
                    if INDEX_19 in self.indexes:
                        ind_19_data = self.get_enrolled_data(
                            index_19, current_year_enrolled_filter)
                        if ind_19_data:
                            dou_enrolled[INDEX_19][age_cat] = ind_19_data

                    if INDEX_19_1 in self.indexes:
                        ind_19_1_data = self.get_enrolled_data(
                            index_19, on_date_enrolled_filter)
                        if ind_19_1_data:
                            dou_enrolled[INDEX_19_1][age_cat] = ind_19_1_data

                    if INDEX_19_2 in self.indexes:
                        ind_19_2_data = self.get_enrolled_data(
                            index_19, next_calendar_year_enrolled)
                        if ind_19_2_data:
                            dou_enrolled[INDEX_19_2][age_cat] = ind_19_2_data

                if INDEX_19_3 in self.indexes:
                    ind_19_3_children_ids = ind_19_3(
                        _dou, report_start_date, age_range=age_cat,
                        prep_direct_func=prep_ind_19_3_queries)
                    ind_19_3_data = ReportQuery.get_child_query(
                        Children.objects.filter(id__in=ind_19_3_children_ids))
                    if ind_19_3_data:
                        dou_enrolled[INDEX_19_3][age_cat] = list(ind_19_3_data)

                if INDEX_29_1 in self.indexes:
                    ind_29_1_data = self.get_enrolled_data(
                        index_29_1, on_date_enrolled_filter)
                    if ind_29_1_data:
                        dou_enrolled[INDEX_29_1][age_cat] = ind_29_1_data

            # Общее количество детей зачисленных в семейные,
            # оздоровительные и компенсирующие группы
            if not calc_filial and INDEX_29 in self.indexes:
                ind_29_data = list(prepared_enrolled_index(
                    query_filter='predictable_decrease_contingent'))

                if ind_29_data:
                    dou_enrolled[INDEX_29][const.ALL] = ind_29_data

            return enrolled

        enrolled_data = _inner(dou, self._report_start_date)

        for filial in self._unit_helper.get_filial(dou).iterator():
            filial_enrolled = _inner(
                filial, self._report_start_date, calc_filial=True)

            self._merge_dicts(enrolled_data, filial_enrolled)

        self._update_cache(enrolled_data, self.ENROLLED)

        return enrolled_data
