from functools import partialmethod
from collections import defaultdict
from collections import namedtuple
from itertools import chain

from django.db.models import CharField
from django.db.models import Value

from kinder.core.children.models import Children

from .constants import DataTypes as DT


# namedtuple для описания строки отчёта с информацией о ребёнке
ReportChildInfo = namedtuple(
    'ReportChildInfo',
    ['unit_name', 'age_cat', 'child_id', 'date_of_birth',
     'fullname', 'data_type', 'id']
)


class ReportQuery:
    """Класс для обработки запросов для отчёта."""

    # Значения values_list для запросов различных моделей
    children_values_list = ('id', 'date_of_birth', 'fullname', 'model', 'id')
    declaration_values_list = pupils_values_list = deduct_values_list = (
        'children_id', 'children__date_of_birth',
        'children__fullname', 'model', 'id'
    )
    direct_values_list = (
        'declaration__children_id', 'declaration__children__date_of_birth',
        'declaration__children__fullname', 'model', 'id',
    )

    # Словарь для соответствия типа данных и values_list для запроса
    type_to_values_list = {
        DT.DIRECT: direct_values_list,
        DT.DECLARATION: declaration_values_list,
        DT.CHILD: children_values_list,
        DT.PUPIL: pupils_values_list,
        DT.DEDUCT: deduct_values_list,
    }

    @classmethod
    def get(cls, data_type, query):
        """Получения запроса с необходимыми данными для отчёта

        :param data_type: Тип данных
        :param query: Запрос

        :return: Запрос с необходимыми данными для запроса
        """
        return query.annotate(
            model=Value(data_type, output_field=CharField())
        ).values_list(*cls.type_to_values_list[data_type])

    # Методы для получения запросов для различных моделей
    get_child_query = partialmethod(get, DT.CHILD)
    get_direct_query = partialmethod(get, DT.DIRECT)
    get_declaration_query = partialmethod(get, DT.DECLARATION)
    get_pupil_query = partialmethod(get, DT.PUPIL)
    get_deduct_query = partialmethod(get, DT.DEDUCT)


def prepare_enrolled_index_queries(
        enrolled_in_group_fact, enrolled_in_group_plan):
    """Функция получения доп. данных для enrolled индекса."""

    return list(chain(
        ReportQuery.get_pupil_query(enrolled_in_group_fact),
        ReportQuery.get_pupil_query(enrolled_in_group_plan),
    ))


def prep_ind_19_3_queries(*queries):
    """Функция получения доп. данных для индекса 19_3."""

    (direct_in_group, direct_count_in_bounded_group,
     pupil_direct_in_bounded_group) = queries

    return list(
        chain(
            direct_in_group.values_list(
                'declaration__children_id',
                'declaration__children__date_of_birth'
            ),
            direct_count_in_bounded_group.values_list(
                'declaration__children_id',
                'declaration__children__date_of_birth'
            ),
            pupil_direct_in_bounded_group.values_list(
                'children_id',
                'children__date_of_birth'
            )
        )
    )


def prep_ind_29_1_queries(*queries):
    """Функция получения доп. данных для индекса 29_1."""

    direct_query, pupil_query = queries
    return set(chain(
        ReportQuery.get_direct_query(direct_query),
        # Как направления считаем и зачисления, дата зачисления по приказу
        # которых больше контрольной даты
        ReportQuery.get_pupil_query(pupil_query),
    ))


def prep_ind_enrolled_data(children_enrollments):
    """Функция получения доп. данных для индекса enrolled."""

    children_data = ReportQuery.get_child_query(
        Children.objects.filter(id__in=children_enrollments.keys()))

    group_enrollments = defaultdict(list)

    for children_id, group_id in children_enrollments.items():

        children_info = {t[0]: t for t in [a for a in children_data]}
        group_enrollments[group_id].append(children_info.get(children_id))

    return group_enrollments
