from collections import defaultdict
import os
import uuid

from simple_report.converter.abstract import FileConverter
from simple_report.report import SpreadsheetReport
from simple_report.xlsx.document import DocumentXLSX

from gisdo import settings as gs_settings
from kinder.core.unit.models import Unit

from .constants import DataTypes
from .constants import INDEXES
from .helpers import ReportChildInfo


class IndexReportAdapter:
    """Адаптер отчёта "Выгрузка детей по показателю/тегу"."""
    def __init__(self, data: list):
        self.data = data

    @staticmethod
    def get_unit_name(unit_id: int) -> str:
        """Получение названия организации по id."""
        return Unit.objects.values_list('name', flat=True).get(id=unit_id)

    def get_data(self) -> dict:
        """Получение преобразованных данных."""
        final_data = defaultdict(list)

        for mo_data in self.data:
            for unit_id, index_data in mo_data.items():
                unit_name = self.get_unit_name(unit_id)

                for index, data in index_data.items():
                    for age_cat, rows in data.items():
                        child_info_list = [
                            ReportChildInfo(unit_name, age_cat, *row)
                            for row in rows
                        ]
                        final_data[index].extend(child_info_list)

        return final_data

    @staticmethod
    def get_row_data(child_info: ReportChildInfo) -> dict:
        """Получения словаря с данными для вставки в отчёт

        :param child_info: Элемент ReportChildInfo с данными о ребёнке

        :return: Словарь с данными для отчёта
        """
        return {
            'doo_name': child_info.unit_name,
            'age_cat': child_info.age_cat,
            'data_type': DataTypes.values.get(child_info.data_type, ''),
            'data_id': child_info.id,
            'child_fullname': child_info.fullname,
        }


class IndexReport:
    """Построитель отчета "Выгрузка детей по показателю/тегу"."""

    def __init__(self, data):
        self.data = data
        self.adapter = IndexReportAdapter(data)

        self.report_name = str(uuid.uuid4())[:16]

        self.template_name = os.path.join(
            os.path.dirname(__file__), "index_report.xlsx")

        self.result_name = os.path.join(
            gs_settings.DOWNLOADS_DIR, self.report_name + ".xlsx")

        self.report = SpreadsheetReport(
            self.template_name, wrapper=DocumentXLSX, type=FileConverter.XLSX)

    def build(self):
        """Построение отчета"""

        header_section = self.report.get_section('header')
        header_section.flush({})

        tag_section = self.report.get_section('tag')
        data_section = self.report.get_section('data')

        data = self.adapter.get_data()

        for index, _ in INDEXES:
            child_info_list = data.get(index)
            if not child_info_list:
                continue

            tag_section.flush({'tag_index': index})

            for child_info in child_info_list:
                row_data = self.adapter.get_row_data(child_info)
                data_section.flush(row_data)

        self.report.build(self.result_name)

        return self.report_name
