import datetime
import traceback

from celery import group
from celery.result import allow_join_result
from django.utils import timezone
from django.utils.functional import cached_property

from gisdo import settings as gs_settings
from gisdo.logger import Logger
from gisdo.models import ScheduleSettings
from kinder.core.async_tasks.tasks import celery_app
from kinder.core.unit.models import Unit

from ..service import ReportService
from ..utils import UnitHelper
from .collector import ReportDataCollector
from .reports import IndexReport


@celery_app.task(queue=gs_settings.CELERY_WORKER_TASK_QUEUE)
def collect_data(mo_id, main_report_unit, indexes):
    """
    Асинхронная задача сбора данных по организации.

    :param mo_id: kinder.core.unit.models.Unit - Муниципальное образование.
    :param main_report_unit: Организация отчета.
    :param indexes: Индексы по которым идет сборка
    """

    mo = Unit.objects.get(id=mo_id)
    unit_helper = UnitHelper(main_report_unit)

    data = None
    exception = None

    report_start_date = datetime.date.today()

    try:
        data = ReportDataCollector(
            mo, unit_helper, report_start_date, indexes).collect()
    except Exception as e:
        Logger.add_record(traceback.format_exc(), level=Logger.ERROR)
        exception = e
    return mo.id, data, exception


@celery_app.task(queue=gs_settings.CELERY_WORKER_TASK_QUEUE)
def build_mo_data(mo_data, **kw):
    """Задача построения отчета"""

    report = IndexReport(mo_data)
    result_name = report.build()

    return result_name


class IndexReportService(ReportService):
    """Сборщик и построитель отчета "Выгрузка детей по показателю/тегу"."""

    def __init__(self, main_unit, indexes, set_progress_func=None, user=None):
        # Перекрыт, для того, чтобы не создавать объект ReportForm
        super(ReportService, self).__init__()

        self._unit_helper = UnitHelper(main_unit)
        self.main_report_unit = main_unit
        self.indexes = indexes

        self._set_progress = set_progress_func or (lambda x: x)
        self._schedule_settings = ScheduleSettings.get_settings()

    @cached_property
    def mo_count(self):
        return len(list(self._unit_helper.get_report_units()))

    def _collect_mo_data(self):
        """Сборка отчетов по каждому МО."""

        # Создаем группу параллельно выполняющихся подзадач
        # Каждая задача отвечает за подсчет показателей отдельного МО
        collect_data_tasks = group(
            collect_data.s(
                mo.id, self.main_report_unit, self.indexes
            ) for mo in self._unit_helper.get_report_units()
        )

        collect_data_tasks_results = collect_data_tasks(
            queue=gs_settings.CELERY_WORKER_TASK_QUEUE
        )

        mo_data_for_report = []
        with allow_join_result():
            for (
                current_mo_num, (mo_id, task_result, exception)
            ) in enumerate(collect_data_tasks_results.iterate(), 1):
                self._set_progress(
                    f'Обработано МО: {current_mo_num} из {self.mo_count}'
                )

                mo = Unit.objects.get(id=mo_id)
                if exception:
                    status_msg = (
                        f'Во время сбора данных по {mo} '
                        f'произошла ошибка: {exception}'
                    )
                else:
                    if self._unit_helper.is_region:
                        self._update_cache(task_result, self.ENROLLED)
                    mo_data_for_report.append(task_result)
                    status_msg = f'Данные по {mo} успешно собраны'

                self._set_progress(
                    values={str(timezone.now()): status_msg}
                )

        return mo_data_for_report

    def _build_report_data(self, mo_data_for_build):
        """Построение данных в xlsx отчет"""

        self._set_progress(
            'Подготовка к формированнию данных отчета по показателям/тегу'
        )

        result_name = None
        try:
            result_name = build_mo_data(mo_data_for_build)
            status_msg = f'Данные успешно сформированы'
        except Exception as e:
            status_msg = (
                f'При формировании данных произошла ошибка: {e}'
            )

        self._set_progress(
            values={str(timezone.now()): status_msg}
        )

        return result_name

    def collect_and_build(self):
        mo_data_for_build = self._collect_mo_data()
        result_name = self._build_report_data(mo_data_for_build)

        return result_name
